from .terms import Psuedocolumn

ColumnValue = Psuedocolumn('COLUMN_VALUE')
ObjectID = Psuedocolumn('OBJECT_ID')
ObjectValue = Psuedocolumn('OBJECT_VALUE')
RowNum = Psuedocolumn('ROWNUM')
RowID = Psuedocolumn('ROWID')
SysDate = Psuedocolumn('SYSDATE')
